//
// C++ Interface: baro
//
// Description: 
//
//
// Author: Thomas Sailer <t.sailer@alumni.ethz.ch>, (C) 2007
//
// Copyright: See COPYING file that comes with this distribution
//
//

#ifndef BARO_H
#define BARO_H

class IcaoAtmosphere {
        public:
                static const float const_g   = 9.80665;                     /* gravity const. */
                static const float const_m   = 28.9644;                     /* Molecular Weight of air */
                static const float const_r   = 8.31432;                     /* gas const. for air */
                static const float const_gmr;                               /* Hydrostatic constant */
		static const float degc_to_kelvin = 273.15;

                IcaoAtmosphere(float qnh = 1013.25, float temp = 288.15);
                void set(float qnh = 1013.25, float temp = 288.15);
                void altitute_to_pressure(float *press, float *temp, float alt);
                void pressure_to_altitute(float *alt, float *temp, float press);
                static float pressure_to_density(float pressure, float temp);
                static float density_to_pressure(float density, float temp);

        private:
                class AtmospericLayer {
                        public:
                                float limit;
                                float lapserate;
                };

                //static const unsigned int nr_layers = sizeof(atmosperic_layers)/sizeof(atmosperic_layers[0])-1;
                static const unsigned int nr_layers = 7;
                static const AtmospericLayer atmosperic_layers[nr_layers+1];
                struct atmospheric_boundary {
                        double t;
                        double p;
                } b[nr_layers+1];

                void compute_boundaries(void);
};

#endif /* BARO_H */
