//
// C++ Interface: maps
//
// Description: 
//
//
// Author: Thomas Sailer <t.sailer@alumni.ethz.ch>, (C) 2007, 2008
//
// Copyright: See COPYING file that comes with this distribution
//
//

#ifndef MAPS_H
#define MAPS_H

#include <limits>
#include <glibmm.h>
#include <gtkmm/drawingarea.h>
//#include <gtkmm.h>
#include <libglademm.h>
#include <sigc++/sigc++.h>

#include "engine.h"
#include "maps.h"

class MapRenderer : public sigc::trackable {
        public:
                typedef enum {
                                drawflags_none = 0,
                                drawflags_topo = 1 << 0,
                                drawflags_terrain = 1 << 1,
                                drawflags_terrain_names = 1 << 2,
                                drawflags_terrain_borders = 1 << 3,
                                drawflags_navaids_vor = 1 << 4,
                                drawflags_navaids_dme = 1 << 5,
                                drawflags_navaids_ndb = 1 << 6,
                                drawflags_navaids = drawflags_navaids_vor | drawflags_navaids_dme | drawflags_navaids_ndb,
                                drawflags_waypoints_low = 1 << 7,
                                drawflags_waypoints_high = 1 << 8,
                                drawflags_waypoints_rnav = 1 << 9,
                                drawflags_waypoints_terminal = 1 << 10,
                                drawflags_waypoints_vfr = 1 << 11,
                                drawflags_waypoints_user = 1 << 12,
                                drawflags_waypoints = drawflags_waypoints_low | drawflags_waypoints_high | drawflags_waypoints_rnav | drawflags_waypoints_terminal | drawflags_waypoints_vfr | drawflags_waypoints_user,
                                drawflags_airports_civ = 1 << 13,
                                drawflags_airports_mil = 1 << 14,
                                drawflags_airports = drawflags_airports_civ | drawflags_airports_mil,
                                drawflags_airports_localizercone = 1 << 15,
                                drawflags_airspaces_ab = 1 << 16,
                                drawflags_airspaces_cd = 1 << 17,
                                drawflags_airspaces_efg = 1 << 18,
                                drawflags_airspaces_specialuse = 1 << 19,
                                drawflags_airspaces = drawflags_airspaces_ab | drawflags_airspaces_cd | drawflags_airspaces_efg | drawflags_airspaces_specialuse,
                                drawflags_airspaces_fill_ground = 1 << 20,
                                drawflags_route = 1 << 21,
                                drawflags_track = 1 << 22,
                                drawflags_track_points = 1 << 23,
                                drawflags_airways_low = 1 << 24,
                                drawflags_airways_high = 1 << 25,
                                drawflags_all = drawflags_topo | drawflags_terrain | drawflags_terrain_names | drawflags_terrain_borders |
                                                drawflags_navaids | drawflags_waypoints | drawflags_airports | drawflags_airports_localizercone |
                                                drawflags_airspaces | drawflags_airspaces_fill_ground |
                                                drawflags_route | drawflags_track | drawflags_track_points |
                                                drawflags_airways_low | drawflags_airways_high
                } DrawFlags;

                template <typename T> class ScreenCoordTemplate {
                        public:
                                ScreenCoordTemplate(T xx = 0, T yy = 0) : x(xx), y(yy) {}
                                T getx(void) const { return x; }
                                T gety(void) const { return y; }
                                void setx(T xx = 0) { x = xx; }
                                void sety(T yy = 0) { y = yy; }
                                bool operator==(const ScreenCoordTemplate &p) const { return x == p.x && y == p.y; }
                                bool operator!=(const ScreenCoordTemplate &p) const { return !operator==(p); }
                                ScreenCoordTemplate operator+(const ScreenCoordTemplate &p) const { return ScreenCoordTemplate(x + p.x, y + p.y); }
                                ScreenCoordTemplate operator-(const ScreenCoordTemplate &p) const { return ScreenCoordTemplate(x - p.x, y - p.y); }
                                ScreenCoordTemplate operator-() const { return ScreenCoordTemplate(-x, -y); }
                                ScreenCoordTemplate& operator+=(const ScreenCoordTemplate &p) { x += p.x; y += p.y; return *this; }
                                ScreenCoordTemplate& operator-=(const ScreenCoordTemplate &p) { x -= p.x; y -= p.y; return *this; }
                                bool operator<(const ScreenCoordTemplate &p) const { return x < p.x && y < p.y; }
                                bool operator<=(const ScreenCoordTemplate &p) const { return x <= p.x && y <= p.y; }
                                bool operator>(const ScreenCoordTemplate &p) const { return x > p.x && y > p.y; }
                                bool operator>=(const ScreenCoordTemplate &p) const { return x >= p.x && y >= p.y; }
                                ScreenCoordTemplate operator*(T z) const { return ScreenCoordTemplate(x * z, y * z); }
                                ScreenCoordTemplate& operator*=(T z) { x *= z; y *= z; return *this; }
                                ScreenCoordTemplate operator/(T z) const { return ScreenCoordTemplate(x / z, y / z); }
                                ScreenCoordTemplate& operator/=(T z) { x /= z; y /= z; return *this; }
                        private:
                                T x, y;
                };
                typedef ScreenCoordTemplate<int> ScreenCoord;
                typedef ScreenCoordTemplate<float> ScreenCoordFloat;

                MapRenderer(Engine& eng, const ScreenCoord& scrsize = ScreenCoord(0, 0), const Point& center = Point(), int alt = 0);
                virtual ~MapRenderer() {}
                virtual void set_route(const FPlanRoute *route) { m_route = route; }
                virtual void set_track(const TracksDb::Track *track) { m_track = track; }
                virtual void redraw_route(void) = 0;
                virtual void redraw_track(void) = 0;
                Point get_center(void) const { return m_center; }
                int get_altitude(void) const { return m_altitude; }
                virtual void set_center(const Point& pt, int alt) { m_center = pt; m_altitude = alt; }
                ScreenCoord get_screensize(void) const { return m_screensize; }
                virtual void set_screensize(const ScreenCoord& scrsize) { m_screensize = scrsize; }
                virtual void set_map_scale(float nmi_per_pixel) = 0;
                virtual float get_map_scale(void) const { return 1; }
                virtual void zoom_out(void) { set_map_scale(get_map_scale() * 2.0); }
                virtual void zoom_in(void) { set_map_scale(get_map_scale() * 0.5); }
                virtual void pan_up(void);
                virtual void pan_down(void);
                virtual void pan_left(void);
                virtual void pan_right(void);
                virtual void set_maxalt(int alt = std::numeric_limits<int>::max()) { }
                virtual int get_maxalt(void) const { return std::numeric_limits<int>::max(); }
                virtual operator DrawFlags() const = 0;
                virtual MapRenderer& operator=(DrawFlags f) = 0;
                virtual MapRenderer& operator&=(DrawFlags f) = 0;
                virtual MapRenderer& operator|=(DrawFlags f) = 0;
                virtual MapRenderer& operator^=(DrawFlags f) = 0;
                virtual void airspaces_changed(void) = 0;
                virtual void airports_changed(void) = 0;
                virtual void navaids_changed(void) = 0;
                virtual void waypoints_changed(void) = 0;
                virtual void airways_changed(void) = 0;
                virtual void draw(Glib::RefPtr<Gdk::Drawable>& drawable, ScreenCoord xy, ScreenCoord wh, ScreenCoord offs = ScreenCoord(0, 0));
                virtual Point screen_coord(const ScreenCoord& sc) = 0;
                virtual Point screen_coord(const ScreenCoordFloat& sc) = 0;
                virtual ScreenCoord coord_screen(const Point& pt) = 0;
                virtual void hide(void) {}
                virtual void show(void) {}
                virtual sigc::connection connect_update(const sigc::slot<void>& slot) = 0;

        protected:
                Point m_center;
                int m_altitude;
                Engine& m_engine;
                const FPlanRoute *m_route;
                const TracksDb::Track *m_track;
                ScreenCoord m_screensize;
};

template<typename T> std::ostream& operator<<(std::ostream& os, const MapRenderer::ScreenCoordTemplate<T>& sc);

class VectorMapRenderer : public MapRenderer {
        public:
                class TimeMeasurement {
                        public:
                                TimeMeasurement(bool start = true);
                                void start(void);
                                void stop(void);
                                operator unsigned int();
                                std::ostream& print(std::ostream& os);

                        private:
                                Glib::TimeVal m_start, m_stop;
                                bool m_running;
                };
                friend std::ostream& operator<<(std::ostream& os, VectorMapRenderer::TimeMeasurement& tm);

        private:
                class DrawThread {
                        public:
                                DrawThread(Engine& eng, int depth, const ScreenCoord& scrsize = ScreenCoord(0, 0), const Point& center = Point(), int alt = 0);
                                virtual ~DrawThread();

                                void set_screensize(const ScreenCoord& scrsize);
                                const ScreenCoord& get_screensize(void) const { return m_screensize; }
                                void set_center(const Point& pt, int alt);
                                const Point& get_center(void) const { return m_center; }
                                virtual void set_map_scale(float nmi_per_pixel);
                                float get_map_scale(void) const { return m_invmapscale * (60.0 / Point::from_deg); }
                                void set_maxalt(int alt = std::numeric_limits<int>::max()) { m_maxalt = alt; }
                                int get_maxalt(void) const { return m_maxalt; }
                                DrawFlags set_drawflags(DrawFlags df);
                                DrawFlags get_drawflags(void) const { return m_drawflags; }
                                virtual void airspaces_changed(void);
                                virtual void airports_changed(void);
                                virtual void navaids_changed(void);
                                virtual void waypoints_changed(void);
                                virtual void airways_changed(void);
                                Cairo::RefPtr<Cairo::ImageSurface> get_surface(void);

                                void hide(void);

                                void draw(Glib::RefPtr<Gdk::Drawable>& drawable, ScreenCoord xy, ScreenCoord wh, ScreenCoord offs = ScreenCoord(0, 0));

                                void draw_scale(Cairo::Context *cr, float target_pixel = 100);
                                void draw_route(Cairo::Context *cr, const FPlanRoute *route, ScreenCoord xy = ScreenCoord(0, 0), ScreenCoord offs = ScreenCoord(0, 0));
                                void draw_track(Cairo::Context *cr, const TracksDb::Track *track, ScreenCoord xy = ScreenCoord(0, 0), ScreenCoord offs = ScreenCoord(0, 0));

                                Point screen_coord(const ScreenCoord& sc);
                                Point screen_coord(const ScreenCoordFloat& sc);
                                ScreenCoord coord_screen(const Point& pt);

                                sigc::connection connect_update(const sigc::slot<void>& slot) { return m_dispatch.connect(slot); }

                        protected:
                                class AirportRoutePoint {
                                        public:
                                                AirportRoutePoint(const Glib::ustring& name = "", const Point& coord = Point(), AirportsDb::Airport::label_placement_t lp = AirportsDb::Airport::label_off);
                                                AirportRoutePoint(const AirportsDb::Airport::VFRRoute::VFRRoutePoint& pt);
                                                bool operator<(const AirportRoutePoint& p) const;
                                                Glib::ustring get_name(void) const { return m_name; }
                                                const Point& get_coord(void) const { return m_coord; }
                                                AirportsDb::Airport::label_placement_t get_label_placement(void) const { return m_labelplacement; }

                                        protected:
                                                Glib::ustring m_name;
                                                Point m_coord;
                                                AirportsDb::Airport::label_placement_t m_labelplacement;
                                };

                                // variables shared by the rendering and the gui threads
                                Engine& m_engine;
                                Point m_center;
                                int m_altitude;
                                int m_depth;
                                ScreenCoord m_screensize;
                                float m_mapscale, m_invmapscale, m_mapscalelon, m_invmapscalelon;
                                DrawFlags m_drawflags;
                                Cairo::RefPtr<Cairo::ImageSurface> m_surface;
                                Point m_surfacecenter;
                                bool m_done;
                                bool m_abort;
                                bool m_hidden;
                                bool m_terminate;
                                bool m_mapscalechange;
                                Glib::Cond m_cond;
                                Glib::Mutex m_mutex;
                                Glib::Thread *m_thread;
                                Glib::Dispatcher m_dispatch;
                                // private variables of the GUI thread
                                // private variables of the rendering thread
                                Point m_current_surfacecenter;
                                ScreenCoord m_current_surfacesize;
                                int m_maxalt;

                                void thread(void);

                                void restart(void);

                                virtual void compute_pixmap(void) = 0;

                                void update_mapscalelon(void);
                                ScreenCoord get_current_pixmap_size(void) const { return m_current_surfacesize; }
                                ScreenCoord get_current_center_coord(void) const;
                                Point current_ctrans(const ScreenCoord& p, const Point& surfacecenter) const;
                                Point current_ctrans(const ScreenCoord& p) const;
                                ScreenCoord current_ctrans(const Point& p) const;
                                Point current_cftrans(const ScreenCoordFloat& p, const Point& surfacecenter) const;
                                Point current_cftrans(const ScreenCoordFloat& p) const;
                                ScreenCoordFloat current_cftrans(const Point& p) const;

                                static Glib::ustring freqstr(uint32_t f = 0);

                                ScreenCoord get_pixmap_size(void) const;
                                ScreenCoord get_center_coord(void) const;
                                Point ctrans(const ScreenCoord& p) const;
                                ScreenCoord ctrans(const Point& p) const;
                                Point cftrans(const ScreenCoordFloat& p) const;
                                ScreenCoordFloat cftrans(const Point& p) const;
                                ScreenCoord get_offset(void) const;
                                ScreenCoordFloat get_offset_float(void) const;
                                float nmi_to_pixel(float nmi) const;

                                void polypath(Cairo::Context *cr, const Polygon& poly, bool closeit = false, const ScreenCoordFloat& offset = ScreenCoordFloat());
                                void polypath(Cairo::Context *cr, const PolygonHole& poly, bool closeit = false, const ScreenCoordFloat& offset = ScreenCoordFloat());
                                void draw_gtopo(Cairo::Context *cr, const Rect& cbbox, const Cairo::RefPtr<Cairo::ImageSurface>& pbuf);
                                void draw_waypoint(Cairo::Context *cr, const Point& pt, const Glib::ustring& str, bool mandatory = true, WaypointsDb::Waypoint::label_placement_t lp = WaypointsDb::Waypoint::label_e);
                                void draw_label(Cairo::Context *cr, const Glib::ustring& text, const Point& pt, NavaidsDb::Navaid::label_placement_t lp = NavaidsDb::Navaid::label_center, int dist = 2);
                                void draw_label(Cairo::Context *cr, const Glib::ustring& text, const ScreenCoord& pos, NavaidsDb::Navaid::label_placement_t lp = NavaidsDb::Navaid::label_center, int dist = 2);
                                void draw_airport_vfrreppt(Cairo::Context *cr, const AirportsDb::Airport& arpt);
                                void draw_navaid(Cairo::Context *cr, const NavaidsDb::Navaid& n, const ScreenCoordFloat& pos);
                };

                class DrawThreadVMap : public DrawThread {
                        public:
                                DrawThreadVMap(Engine& eng, int depth, const ScreenCoord& scrsize = ScreenCoord(0, 0), const Point& center = Point(), int alt = 0);
                                virtual ~DrawThreadVMap();

                                virtual void airspaces_changed(void);
                                virtual void airports_changed(void);
                                virtual void navaids_changed(void);
                                virtual void waypoints_changed(void);
                                virtual void airways_changed(void);

                                inline void draw(Glib::RefPtr<Gdk::Drawable>& drawable, ScreenCoord xy, ScreenCoord wh, ScreenCoord offs = ScreenCoord(0, 0)) { DrawThread::draw(drawable, xy, wh, offs); }

                        protected:
                                class mapel_render_comp : public std::binary_function<const MapelementsDb::Mapelement&, const MapelementsDb::Mapelement&, bool> {
                                        public:
                                                mapel_render_comp(void) {}
                                                bool operator()(const MapelementsDb::Mapelement&, const MapelementsDb::Mapelement&) const;
                                        private:
                                                Point pt;
                                };

                                static const float aiport_loccone_centerlength = 5.5;
                                static const float aiport_loccone_sidelength = 6.0;
                                static const float aiport_loccone_sideangle = 5.0;

                                // private variables of the rendering thread
                                DrawFlags m_current_drawflags;
                                Rect m_current_surfacerect;
                                Rect m_current_dbrect;
                                Glib::RefPtr<Engine::ElevationMapCairoResult> m_topocairoquery;
                                Glib::RefPtr<Engine::MapelementResult> m_mapelementquery;
                                Glib::RefPtr<Engine::NavaidResult> m_navaidquery;
                                Glib::RefPtr<Engine::WaypointResult> m_waypointquery;
                                Glib::RefPtr<Engine::AirportResult> m_airportquery;
                                Glib::RefPtr<Engine::AirspaceResult> m_airspacequery;
                                Glib::RefPtr<Engine::AirwayResult> m_airwayquery;
                                bool m_abort_cache;

                                bool check_abort(void);
                                void compute_pixmap(void);
                                void async_cancel(void);
                                void async_done(void);

                                void draw(Cairo::Context *cr, const MapelementsDb::Mapelement& mapel);
                                void draw_text(Cairo::Context *cr, const MapelementsDb::Mapelement& mapel);
                                void draw(Cairo::Context *cr, const std::vector<NavaidsDb::Navaid>& navaids);
                                void draw(Cairo::Context *cr, const std::vector<AirspacesDb::Airspace>& airspaces);
                                void draw(Cairo::Context *cr, const std::vector<AirportsDb::Airport>& airports);
                                void draw(Cairo::Context *cr, const std::vector<WaypointsDb::Waypoint>& waypoints);
                                void draw(Cairo::Context *cr, const std::vector<AirwaysDb::Airway>& airways);
                                static DrawFlags drawflags(const MapelementsDb::Mapelement& mapel) { return drawflags_terrain; }
                                static bool is_ground(const AirspacesDb::Airspace& a);
                };

                class DrawThreadTerrain : public DrawThread {
                        public:
                                DrawThreadTerrain(Engine& eng, int depth, const ScreenCoord& scrsize = ScreenCoord(0, 0), const Point& center = Point(), int alt = 0);
                                virtual ~DrawThreadTerrain();

                                virtual void airspaces_changed(void);
                                virtual void airports_changed(void);
                                virtual void navaids_changed(void);
                                virtual void waypoints_changed(void);
                                virtual void airways_changed(void);

                                inline void draw(Glib::RefPtr<Gdk::Drawable>& drawable, ScreenCoord xy, ScreenCoord wh, ScreenCoord offs = ScreenCoord(0, 0)) { DrawThread::draw(drawable, xy, wh, offs); }

                        protected:
                                Rect m_current_surfacerect;
                                Rect m_current_dbrect;
                                int m_current_surfacealtitude;
                                Glib::RefPtr<Engine::ElevationMapResult> m_topoquery;
                                bool m_abort_cache;

                                bool check_abort(void);
                                void compute_pixmap(void);
                                void async_cancel(void);
                                void async_done(void);
                };

                class DrawThreadAirportDiagram : public DrawThread {
                        public:
                                DrawThreadAirportDiagram(Engine& eng, int depth, const ScreenCoord& scrsize = ScreenCoord(0, 0), const Point& center = Point(), int alt = 0);
                                virtual ~DrawThreadAirportDiagram();

                                virtual void set_map_scale(float nmi_per_pixel);

                                virtual void airspaces_changed(void);
                                virtual void airports_changed(void);
                                virtual void navaids_changed(void);
                                virtual void waypoints_changed(void);
                                virtual void airways_changed(void);

                                inline void draw(Glib::RefPtr<Gdk::Drawable>& drawable, ScreenCoord xy, ScreenCoord wh, ScreenCoord offs = ScreenCoord(0, 0)) { DrawThread::draw(drawable, xy, wh, offs); }

                        protected:
                                class LabelPart : public Glib::ustring {
                                        public:
                                                LabelPart(void) : m_fgr(0.0), m_fgg(0.0), m_fgb(0.0), m_bgr(1.0), m_bgg(1.0), m_bgb(0.0) {}
                                                void set_fg(double r, double g, double b) { m_fgr = r; m_fgg = g; m_fgb = b; }
                                                void set_bg(double r, double g, double b) { m_bgr = r; m_bgg = g; m_bgb = b; }
                                                void set_fg(Cairo::Context *cr) const { cr->set_source_rgb(m_fgr, m_fgg, m_fgb); }
                                                void set_bg(Cairo::Context *cr) const { cr->set_source_rgb(m_bgr, m_bgg, m_bgb); }

                                        private:
                                                double m_fgr, m_fgg, m_fgb;
                                                double m_bgr, m_bgg, m_bgb;
                                };

                                Rect m_current_surfacerect;
                                Rect m_current_dbrect;
                                Glib::RefPtr<Engine::AirportResult> m_airportquery;
                                Glib::RefPtr<Engine::NavaidResult> m_navaidquery;
                                Glib::RefPtr<Engine::WaypointResult> m_waypointquery;
                                bool m_abort_cache;

                                bool check_abort(void);
                                void compute_pixmap(void);
                                void async_cancel(void);
                                void async_done(void);

                                void draw(Cairo::Context *cr, const AirportsDb::Airport& arpt);
                                void draw(Cairo::Context *cr, const NavaidsDb::Navaid& n);
                                void draw(Cairo::Context *cr, const WaypointsDb::Waypoint& w);
                                void polyline_to_path(Cairo::Context *cr, const AirportsDb::Airport::Polyline& pl);
                                static void draw_taxiwaysign(Cairo::Context *cr, const Glib::ustring& text);
                };

        public:
                typedef enum {
                        renderer_vmap,
                        renderer_terrain,
                        renderer_airportdiagram
                } renderer_t;

                VectorMapRenderer(Engine& eng, int depth, const ScreenCoord& scrsize = ScreenCoord(0, 0), const Point& center = Point(), renderer_t renderer = renderer_vmap);
                virtual ~VectorMapRenderer();
                virtual void redraw_route(void);
                virtual void redraw_track(void);
                virtual void set_center(const Point& pt, int alt);
                virtual void set_screensize(const ScreenCoord& scrsize);
                virtual void set_map_scale(float nmi_per_pixel) { if (m_thread) m_thread->set_map_scale(nmi_per_pixel); }
                virtual float get_map_scale(void) const { return m_thread ? m_thread->get_map_scale() : 0.1; }
                virtual void set_maxalt(int alt = std::numeric_limits<int>::max()) { if (m_thread) m_thread->set_maxalt(alt); }
                virtual int get_maxalt(void) const { return m_thread ? m_thread->get_maxalt() : std::numeric_limits<int>::max(); }
                virtual operator DrawFlags() const { return m_drawflags; }
                virtual MapRenderer& operator=(DrawFlags f);
                virtual MapRenderer& operator&=(DrawFlags f);
                virtual MapRenderer& operator|=(DrawFlags f);
                virtual MapRenderer& operator^=(DrawFlags f);
                virtual void airspaces_changed(void);
                virtual void airports_changed(void);
                virtual void navaids_changed(void);
                virtual void waypoints_changed(void);
                virtual void airways_changed(void);
                virtual void draw(Glib::RefPtr<Gdk::Drawable>& drawable, ScreenCoord xy, ScreenCoord wh, ScreenCoord offs = ScreenCoord(0, 0));
                virtual Point screen_coord(const ScreenCoord& sc) { return m_thread ? m_thread->screen_coord(sc) : Point(); }
                virtual Point screen_coord(const ScreenCoordFloat& sc) { return m_thread ? m_thread->screen_coord(sc) : Point(); }
                virtual ScreenCoord coord_screen(const Point& pt) { return m_thread ? m_thread->coord_screen(pt) : ScreenCoord(0, 0); }
                virtual void hide(void) { if (m_thread) m_thread->hide(); }
                virtual sigc::connection connect_update(const sigc::slot<void>& slot) { return m_thread ? m_thread->connect_update(slot) : sigc::connection(); }

        protected:
                DrawThread *m_thread;
                DrawFlags m_drawflags;
};

inline MapRenderer::DrawFlags operator|(MapRenderer::DrawFlags x, MapRenderer::DrawFlags y) { return (MapRenderer::DrawFlags)((unsigned int)x | (unsigned int)y); }
inline MapRenderer::DrawFlags operator&(MapRenderer::DrawFlags x, MapRenderer::DrawFlags y) { return (MapRenderer::DrawFlags)((unsigned int)x & (unsigned int)y); }
inline MapRenderer::DrawFlags operator^(MapRenderer::DrawFlags x, MapRenderer::DrawFlags y) { return (MapRenderer::DrawFlags)((unsigned int)x ^ (unsigned int)y); }
inline MapRenderer::DrawFlags operator~(MapRenderer::DrawFlags x){ return MapRenderer::drawflags_all ^ x; }
inline MapRenderer::DrawFlags& operator|=(MapRenderer::DrawFlags& x, MapRenderer::DrawFlags y) { x = x | y; return x; }
inline MapRenderer::DrawFlags& operator&=(MapRenderer::DrawFlags& x, MapRenderer::DrawFlags y) { x = x & y; return x; }
inline MapRenderer::DrawFlags& operator^=(MapRenderer::DrawFlags& x, MapRenderer::DrawFlags y) { x = x ^ y; return x; }

inline std::ostream& operator<<(std::ostream& os, VectorMapRenderer::TimeMeasurement& tm) { return tm.print(os); }

class VectorMapArea : public Gtk::DrawingArea {
        public:
                explicit VectorMapArea(BaseObjectType *castitem, const Glib::RefPtr<Gnome::Glade::Xml>& refxml);
                virtual ~VectorMapArea();

                typedef enum {
                        cursor_invalid,
                        cursor_mouse,
                        cursor_set
                } cursor_validity_t;

                typedef enum {
                        renderer_none,
                        renderer_vmap,
                        renderer_terrain,
                        renderer_airportdiagram
                } renderer_t;

                void set_engine(Engine& eng);
                void set_renderer(renderer_t render);
                renderer_t get_renderer(void) const;
                void set_route(const FPlanRoute& route);
                void set_track(const TracksDb::Track *track);
                void redraw_route(void);
                void redraw_track(void);
                virtual void set_center(const Point& pt, int alt);
                virtual Point get_center(void) const;
                virtual int get_altitude(void) const;
                void set_map_scale(float nmi_per_pixel);
                float get_map_scale(void) const;
                void zoom_out(void);
                void zoom_in(void);
                void pan_up(void);
                void pan_down(void);
                void pan_left(void);
                void pan_right(void);
                void set_dragbutton(unsigned int dragbutton);
                bool is_cursor_valid(void) const { return m_cursorvalid; }
                Point get_cursor(void) const { return m_cursorvalid ? m_cursor : Point(); }
                void set_cursor(const Point& pt);
                Rect get_cursor_rect(void) const;
                void invalidate_cursor(void);
                void set_cursor_angle(float angle);
                float get_cursor_angle(void) const { return m_cursorangle * (1.0f / 64.0f); }
                bool is_cursor_angle_valid(void) const { return m_cursoranglevalid; }
                void invalidate_cursor_angle(void);
                virtual operator MapRenderer::DrawFlags() const;
                virtual VectorMapArea& operator=(MapRenderer::DrawFlags f);
                virtual VectorMapArea& operator&=(MapRenderer::DrawFlags f);
                virtual VectorMapArea& operator|=(MapRenderer::DrawFlags f);
                virtual VectorMapArea& operator^=(MapRenderer::DrawFlags f);
                void airspaces_changed(void);
                void airports_changed(void);
                void navaids_changed(void);
                void waypoints_changed(void);
                void airways_changed(void);
                sigc::signal<void,Point,cursor_validity_t> signal_cursor(void) { return m_signal_cursor; }
                sigc::signal<void,Point> signal_pointer(void) { return m_signal_pointer; }

        private:
                bool on_my_expose_event(GdkEventExpose *event);
                void on_my_size_allocate(Gtk::Allocation& allocation);
                void on_my_show(void);
                void on_my_hide(void);
                static unsigned int button_translate(GdkEventButton* event);
                bool on_my_button_press_event(GdkEventButton* event);
                bool on_my_button_release_event(GdkEventButton* event);
                bool on_my_motion_notify_event(GdkEventMotion* event);

        protected:
                void draw_pixmap(GdkEventExpose *event);

                void redraw(void) { queue_draw(); }
                void renderer_update(void);
                void renderer_alloc(void);

        protected:
                Engine *m_engine;
                MapRenderer *m_renderer;
                renderer_t m_rendertype;
                unsigned int m_dragbutton;
                MapRenderer::ScreenCoordFloat m_dragstart;
                MapRenderer::ScreenCoord m_dragoffset;
                bool m_draginprogress, m_cursorvalid, m_cursoranglevalid;
                Point m_cursor;
                int m_cursorangle;
                sigc::signal<void,Point,cursor_validity_t> m_signal_cursor;
                sigc::signal<void,Point> m_signal_pointer;
                // Caching
                Point m_center;
                int m_altitude;
                float m_mapscale;
                const FPlanRoute *m_route;
                const TracksDb::Track *m_track;
                MapRenderer::DrawFlags m_drawflags;
};

#endif /* MAPS_H */
